//-------------------------------------------------------------- 2016-12-03 -----
// VFD Clock
//
//  Ver1.00
//-------------------------------------------------------------------------------
#include <Wire.h>
#include <MsTimer2.h>
#include <EEPROM.h>
#include <Time.h>

//---- PORT -------------------------------
#define A6812_CLOCK      10    // VFD 20-bit serial-input latched source driver
#define A6812_DATA        8
#define A6812_BLANK       9
#define A6812_STROBE     A0
#define LED               6   // red LED
#define VOLTAGE_DETECTOR A1   // MCP100-475DI
#define RTC_OUT           2   // DS1307
#define BUZZER            5   // BUZZER port
#define SW1               7   // SW1 port
#define SW2               4   // SW2 port
#define SW3              A3   // SW3 port
#define SW4              A2   // SW4 port
#define DC_DC_DISABLE     3   // 18V DC-DC

//---- PUSH SWITCH ------------------------
#define SW_ON  LOW
#define SW_OFF HIGH
#define SW_NO1  1
#define SW_NO2  2
#define SW_NO3  3
#define SW_NO4  4
#define SW_DEFAULT_ON_TIME   20  // 20[msec]
#define SW_DEFAULT_OFF_TIME  50  // 50[msec]

//---- DS1307 -----------------------------
#define CLOCK_HALT             0b10000000
#define CLOCK_START            0b00000000
#define OUT_1Hz                0b00010000
#define OUT_4096HZ             0b00010001
#define OUT_8192HZ             0b00010010
#define OUT_32768HZ            0b00010011
#define OUT_OFF                0b00000000
#define DS1307_DEVICE_ADDRESS  0x68
#define DS1307_READ            0b00000001
#define DS1307_WRITE           0b00000000
#define HOURS_AM   1
#define HOURS_PM   0
#define ADDRESS_SECONDS    0
#define ADDRESS_MINUTES    1
#define ADDRESS_HOURS      2
#define ADDRESS_DAY        3
#define ADDRESS_DATE       4
#define ADDRESS_MONTH      5
#define ADDRESS_YEAR       6
#define ADDRESS_CONTROL    7

#define OFFSET_ADDRESS_POWER_OFF_TIME       0x08
#define OFFSET_ADDRESS_CALIBRATION_COUNTER  0x0C

//----  EEPROM address -----------
#define EEPADDR_ALARM_HOURS       0   // 1 byte    default value = 7
#define EEPADDR_ALARM_MINUTES     1   // 1 byte    default value = 0
#define EEPADDR_ALARM_ENABLE      2   // 1 byte    default value = 0 = disable
#define EEPADDR_CALIBRATION_SIGN  3   // 1 byte    '+': fast  '-': behind
#define EEPADDR_CALIBRATION_TIME  4   // 1 byte    XX sec / day fast or behind
#define EEPADDR_DISPLAY_FORMAT    5   // 1 byte    default value = DISPLAY_24HOUR
#define EEPADDR_TIMER_SEC         6   // 1 byte    kitchen timer second
#define EEPADDR_TIMER_MIN         7   // 1 byte    kitchen timer minute
#define EEPADDR_CRC_LOW          0x10   // 1 byte    
#define EEPADDR_CRC_HIGH         0x11   // 1 byte    

#define DEFAULT_CALIBRATION_TIME  8

#define DISPLAY_12HOUR       0
#define DISPLAY_24HOUR       1
#define DISPLAY_ALARM        2
#define DISPLAY_DATE         3
#define DISPLAY_TIMER        4
#define DISPLAY_CALIBRATION  5

#define EEPROM_DATA_SIZE          8

//---- Alarm ---------------------
#define ALARM_ENABLE   0
#define ALARM_DISABLE  1
#define ALARM_DEFAULT_HOUR    7
#define ALARM_DEFAULT_MINUTE  0

//---- Gloval Variable -------------------------------
time_t  alarm_time;
boolean alarm_enable;
byte    display_format;  // 12hour / 24hour
byte    tick_flag;
// byte green_bulb;

boolean sw1_enable;
boolean sw2_enable;
boolean sw3_enable;
boolean sw4_enable;
word sw1_disable_time;
word sw2_disable_time;
word sw3_disable_time;
word sw4_disable_time;
word sw1_disable_counter;
word sw2_disable_counter;
word sw3_disable_counter;
word sw4_disable_counter;

boolean alarm_beep_enable;
byte    alarm_beep_table_index;
word    alarm_beep_counter;
//                                      on   off  on   off  on   off  on   off
const word alarm_beep_time_table[8] = { 100, 100, 100, 100, 100, 100, 100, 400 };

char    gps_buffer[100];
byte    gps_buffer_index;
boolean gps_buffer_full;

char serial_write_buffer[100];

long    cal_counter;
long    cal_set;
byte    cal_time;
byte    cal_sign;

word digit6;
word digit5;
word digit4;
word digit3;
word digit2;
word digit1;

boolean kitchen_counter_init;
boolean kitchen_counter_tick;

time_t current_time;

//---- small function -------------
#define buzzer_on()    analogWrite( BUZZER, 128 )
#define buzzer_off()   digitalWrite( BUZZER, LOW )
#define led_on()       digitalWrite( LED, HIGH )
#define led_off()      digitalWrite( LED, LOW  )

//===============================================================================
// rtc interrupt
//===============================================================================
void rtcEvent()
{
  if ( digitalRead( RTC_OUT ) == 0 ) {
    tick_flag = true;
    current_time++;
  }
}

//===============================================================================
// timer interrupt        1[msec] period
//===============================================================================
void timerEvent()
{
  static word timer = 0;

  //-------- kitchen counter -------------------------------------------
  if ( kitchen_counter_init ) {
    timer = 1000;
    kitchen_counter_init = false;
  } else {
    if ( timer > 0 ) {
      timer--;
    }
    if ( timer == 0 ) {
      kitchen_counter_tick = true;
      timer = 1000;
    }
  }

  //-------- Switch test -------------------------------------------
  //------------- SW1 ------------------
  if ( digitalRead(SW1) == SW_ON ) {
    sw1_disable_counter = sw1_disable_time;
  } else {
    if ( sw1_disable_counter > 0 ) {
      sw1_disable_counter--;
    } else {
      sw1_enable = true;
    }
  }

  //------------- SW2 ------------------
  if ( digitalRead(SW2) == SW_ON ) {
    sw2_disable_counter = sw2_disable_time;
  } else {
    if ( sw2_disable_counter > 0 ) {
      sw2_disable_counter--;
    } else {
      sw2_enable = true;
    }
  }

  //------------- SW3 ------------------
  if ( digitalRead(SW3) == SW_ON ) {
    sw3_disable_counter = sw3_disable_time;
  } else {
    if ( sw3_disable_counter > 0 ) {
      sw3_disable_counter--;
    } else {
      sw3_enable = true;
    }
  }

  //------------- SW4 ------------------
  if ( digitalRead(SW4) == SW_ON ) {
    sw4_disable_counter = sw4_disable_time;
  } else {
    if ( sw4_disable_counter > 0 ) {
      sw4_disable_counter--;
    } else {
      sw4_enable = true;
    }
  }

  //-------- alarm beep ------------------------
  if( alarm_beep_enable ) {
    if( alarm_beep_counter == 0 ) {
      if( alarm_beep_table_index % 2 == 0 ) {
        buzzer_on();
      } else {
        buzzer_off();
      }
      alarm_beep_counter = alarm_beep_time_table[ alarm_beep_table_index ];
      if( alarm_beep_table_index == 7 ) {
        alarm_beep_table_index = 0;
      } else {
        alarm_beep_table_index++;
      }
    } else {
      if( alarm_beep_counter > 0 ) {
        alarm_beep_counter--;
      }
    }
  }
}

//-------------------------------------------------------------------------------
// serial interrupt
//-------------------------------------------------------------------------------
void serialEvent()
{
  byte recieve_data;

  //------- GPS -------------------------------------------------
  if( Serial.available() > 0 ) {
    recieve_data = Serial.read();
    if( recieve_data == '$') {
      gps_buffer_index = 0;
    }
    gps_buffer[gps_buffer_index] = recieve_data;
    if( (gps_buffer_index == 64) or ( recieve_data == 0x0D ) or ( recieve_data == 0x0A ) ) {
      gps_buffer_full = true;
    } else {
      gps_buffer_index++;
    }
  }
}

//-------------------------------------------------------------------------------
// beep
//-------------------------------------------------------------------------------
void beep( word time )
{
  word count;

  buzzer_on();
  for ( count = 0 ; count <= time ; count++ ) {
    delay( 1 );
  }
  buzzer_off();
}

//-------------------------------------------------------------------------------
// beep beep
//-------------------------------------------------------------------------------
void beep2( byte count )
{
  byte counter;

  for ( counter = 0 ; counter < count ; counter++ ) {
    buzzer_on();
    delay(100);
    buzzer_off();
    delay(100);
  }
}

//-------------------------------------------------------------------------------
// error beep
//-------------------------------------------------------------------------------
void beep_error()
{
  byte counter;

  for ( counter = 0 ; counter < 5 ; counter++ ) {
    buzzer_on();
    delay(30);
    buzzer_off();
    delay(30);
  }
}

//-------------------------------------------------------------------------------
// convert bcd <-> decimal
//-------------------------------------------------------------------------------
byte bcd2dec(byte val)
{
  return ( (val / 16 * 10) + (val % 16) );
}

byte dec2bcd(byte val)
{
  return ( (val / 10 * 16) + (val % 10) );
}

//-------------------------------------------------------------------------------
// Real Time Clock write
//-------------------------------------------------------------------------------
void write_ds1307(byte address, byte w_data )
{
  Wire.beginTransmission( DS1307_DEVICE_ADDRESS );
  Wire.write(address);
  Wire.write(w_data);
  Wire.endTransmission();
}

//-------------------------------------------------------------------------------
// Real Time Clock read
//-------------------------------------------------------------------------------
byte read_ds1307( byte address )
{
  Wire.beginTransmission( DS1307_DEVICE_ADDRESS );
  Wire.write(address);
  Wire.endTransmission();
  Wire.requestFrom( DS1307_DEVICE_ADDRESS , 1 ) ;
  return Wire.read();
}

//-------------------------------------------------------------------------------
// Real Time Clock read CLOCK HALT
//-------------------------------------------------------------------------------
boolean ds1307_get_halt()
{
  byte sec;

  sec = read_ds1307( ADDRESS_SECONDS );
  if ((sec & CLOCK_HALT) == CLOCK_HALT ) {
    return true;
  } else {
    return false;
  }
}

//-------------------------------------------------------------------------------
// Real Time Clock read minutes
//-------------------------------------------------------------------------------
boolean get_alarm_enable()
{
  if ( EEPROM.read( EEPADDR_ALARM_ENABLE ) == ALARM_ENABLE ) {
    led_on();
    return true;
  } else {
    led_off();
    return false;
  }
}

//-------------------------------------------------------------------------------
// Real Time Clock read minutes
//-------------------------------------------------------------------------------
void put_alarm_enable( boolean enable )
{
  if ( enable ) {
    EEPROM.write( EEPADDR_ALARM_ENABLE, ALARM_ENABLE );     // enable
    led_on();
  } else {
    EEPROM.write( EEPADDR_ALARM_ENABLE, ALARM_DISABLE );    // disable
    led_off();
  }
  crc_eeprom_write();
}

//-------------------------------------------------------------------------------
// Real Time Clock get time
//-------------------------------------------------------------------------------
time_t ds1307_gettime()
{
  tmElements_t tm;
  time_t temp;

  Wire.beginTransmission( DS1307_DEVICE_ADDRESS );
  Wire.write( 0 );
  Wire.endTransmission();
  Wire.requestFrom( DS1307_DEVICE_ADDRESS , 7 ) ;

  tm.Second = bcd2dec(Wire.read() & 0x7f);
  tm.Minute = bcd2dec(Wire.read() );
  tm.Hour   = bcd2dec(Wire.read() & 0x3f);  // mask assumes 24hr clock
  tm.Wday   = bcd2dec(Wire.read() );        // not use
  tm.Day    = bcd2dec(Wire.read() );
  tm.Month  = bcd2dec(Wire.read() );
  tm.Year   = bcd2dec(Wire.read()) + 30;

  return makeTime( tm );
}

//-------------------------------------------------------------------------------
// Real Time Clock set time
//-------------------------------------------------------------------------------
void ds1307_settime(
  byte year,
  byte month,
  byte date,
  byte hour,     // 0 to 23
  byte minutes,
  byte seconds )
{
  write_ds1307(ADDRESS_SECONDS, CLOCK_HALT       );
  write_ds1307(ADDRESS_YEAR,    dec2bcd(year)    );
  write_ds1307(ADDRESS_MONTH,   dec2bcd(month)   );
  write_ds1307(ADDRESS_DATE,    dec2bcd(date)    );
  write_ds1307(ADDRESS_HOURS,   dec2bcd(hour)    );
  write_ds1307(ADDRESS_MINUTES, dec2bcd(minutes) );
  write_ds1307(ADDRESS_SECONDS, dec2bcd(seconds) );
  cal_counter = 0;
}

//-------------------------------------------------------------------------------
// power off time -> Real Time Clock RAM 
//-------------------------------------------------------------------------------
void ds1307_write_ram_long(byte offset_address, long long_data)
{
  union ulong_convert {
    time_t tm;
    byte data[4];
  } convert_data;
  
  convert_data.tm = long_data;
  Wire.beginTransmission( DS1307_DEVICE_ADDRESS );
  Wire.write( offset_address );
  for( byte index = 0; index < 4 ; index++ ) {
    Wire.write(convert_data.data[index]);
  }
  Wire.endTransmission();    
}

//-------------------------------------------------------------------------------
// return Real Time Clock RAM 
//-------------------------------------------------------------------------------
time_t ds1307_read_ram_long(byte offset_address)
{
  union ulong_convert {
    time_t tm;
    byte data[4];
  } convert_data;
  
  Wire.beginTransmission( DS1307_DEVICE_ADDRESS );
  Wire.write( offset_address );
  Wire.endTransmission();
  Wire.requestFrom( DS1307_DEVICE_ADDRESS, 4 );
  for( byte index = 0; index < 4 ; index++ ) {
    convert_data.data[index] = Wire.read();
  }
  
  return convert_data.tm;
}

//-------------------------------------------------------------------------------
// set time
//-------------------------------------------------------------------------------
void settime( time_t tm )
{
  byte yr;

  yr = year(tm) % 100;
  ds1307_settime( yr, month(tm), day(tm), hour(tm), minute(tm), second(tm) );
}

//-------------------------------------------------------------------------------
// push switch ON test
//-------------------------------------------------------------------------------
boolean sw_on_test( byte sw_no, word sw_on_time, word sw_off_time )
{
  word sw1_on_counter;
  word sw2_on_counter;
  word sw3_on_counter;
  word sw4_on_counter;

  sw1_on_counter = 0;
  sw2_on_counter = 0;
  sw3_on_counter = 0;
  sw4_on_counter = 0;
  switch ( sw_no ) {
    case 1:
      while ( (digitalRead(SW1) == SW_ON) && (sw1_enable == true) ) {
        delay(1);
        if ( sw1_on_counter > sw_on_time ) {
          sw1_disable_time = sw_off_time;
          sw1_enable = false;
          return true;
        }
        sw1_on_counter++;
      }
      break;
    case 2:
      while ( (digitalRead(SW2) == SW_ON) && (sw2_enable == true) ) {
        delay(1);
        if ( sw2_on_counter > sw_on_time ) {
          sw2_disable_time = sw_off_time;
          sw2_enable = false;
          return true;
        }
        sw2_on_counter++;
      }
      break;
    case 3:
      while ( (digitalRead(SW3) == SW_ON) && (sw3_enable == true) ) {
        delay(1);
        if ( sw3_on_counter > sw_on_time ) {
          sw3_disable_time = sw_off_time;
          sw3_enable = false;
          return true;
        }
        sw3_on_counter++;
      }
      break;
    case 4:
      while ( (digitalRead(SW4) == SW_ON) && (sw4_enable == true) ) {
        delay(1);
        if ( sw4_on_counter > sw_on_time ) {
          sw4_disable_time = sw_off_time;
          sw4_enable = false;
          return true;
        }
        sw4_on_counter++;
      }
      break;
  }

  return false;
}

//-------------------------------------------------------------------------------
// alarm beep on
//-------------------------------------------------------------------------------
void alarm_beep_on()
{
  alarm_beep_counter = 0;
  alarm_beep_table_index = 0;
  alarm_beep_enable = true;
}

//-------------------------------------------------------------------------------
// alarm beep off
//-------------------------------------------------------------------------------
void alarm_beep_off()
{
  alarm_beep_enable = false;
  digitalWrite( BUZZER, LOW );
}

#define Digit_0     0x003F
#define Digit_1     0x0006
#define Digit_2     0x005B
#define Digit_3     0x004F
#define Digit_4     0x00E6
#define Digit_5     0x006D
#define Digit_6     0x007D
#define Digit_7     0x0007
#define Digit_8     0x007F
#define Digit_9     0x006F
#define Digit_E     0x0079
#define Digit_F     0x0071
#define Digit_U     0x003E
#define Digit_u     0x001C
#define Digit_L     0x0038
#define Digit_C     0x0039
#define Digit_P     0x0073
#define Digit_Y     0x006E
#define Digit_A     0x0077
#define Digit_r     0x0050
#define Digit_n     0x0054
#define Digit_d     0x005E
#define Digit_t     0x0078
#define Digit_G     0x003D
#define Digit_b     0x007C
#define Digit_i     0x0010
#define Digit_l     0x0006
#define Digit_NON   0x0000
#define Digit_o     0x005C
#define Digit_S     0x006D
#define Digit_h     0x0074
#define Digit_H     0x0076
#define Digit_c     0x0058
#define Digit_US    0x0008
#define Digit_MINUS 0x0040
#define Digit_PLUS  0x00C6
#define Digit_DP    0x0100

word decode_number( byte number )
{
  switch ( number ) {
    case 0: return Digit_0; break;
    case 1: return Digit_1; break;
    case 2: return Digit_2; break;
    case 3: return Digit_3; break;
    case 4: return Digit_4; break;
    case 5: return Digit_5; break;
    case 6: return Digit_6; break;
    case 7: return Digit_7; break;
    case 8: return Digit_8; break;
    case 9: return Digit_9; break;
  }
}

void A6812_send( word bit1to10, word bit11to20 )
{
  for ( byte counter = 1 ; counter <= 10 ; counter++ ) {
    if ( (bit11to20 & 0x0200) == 0x0200 ) {
      digitalWrite( A6812_DATA, HIGH );
    } else {
      digitalWrite( A6812_DATA, LOW );
    }
    digitalWrite( A6812_CLOCK, HIGH );
    digitalWrite( A6812_CLOCK, LOW  );
    bit11to20 <<= 1;
  }
  for ( byte counter = 1 ; counter <= 10 ; counter++ ) {
    if ( (bit1to10 & 0x0200) != 0x0000 ) {
      digitalWrite( A6812_DATA, HIGH );
    } else {
      digitalWrite( A6812_DATA, LOW );
    }
    digitalWrite( A6812_CLOCK, HIGH );
    digitalWrite( A6812_CLOCK, LOW  );
    bit1to10 <<= 1;
  }
}

void display_char( byte chr6, byte chr5, byte chr4, byte chr3, byte chr2, byte chr1 )
{
  A6812_send( chr6, chr5 );
  A6812_send( chr4, chr3 );
  A6812_send( chr2, chr1 );

  digitalWrite( A6812_STROBE, HIGH );
  digitalWrite( A6812_STROBE, LOW  );
  digitalWrite( A6812_BLANK,  LOW  );
}

void display_number( byte dig6, byte dig5, byte dig4, byte dig3, byte dig2, byte dig1 )
{
  display_char( decode_number( dig6 ),  decode_number( dig5 ) | Digit_DP,
                decode_number( dig4 ),  decode_number( dig3 ) | Digit_DP,
                decode_number( dig2 ),  decode_number( dig1 ) | Digit_DP );
}

void display_refresh()
{
  A6812_send( digit6, digit5 );
  A6812_send( digit4, digit3 );
  A6812_send( digit2, digit1 );

  digitalWrite( A6812_STROBE, HIGH );
  digitalWrite( A6812_STROBE, LOW  );
  digitalWrite( A6812_BLANK,  LOW  );
}

void display_clear()
{
  A6812_send( 0, 0 );
  A6812_send( 0, 0 );
  A6812_send( 0, 0 );
  digitalWrite( A6812_STROBE, HIGH );
  digitalWrite( A6812_STROBE, LOW  );
  digitalWrite( A6812_BLANK,  LOW  );
}

//-------------------------------------------------------------------------------
//  display time
//-------------------------------------------------------------------------------
void display_time( time_t tm, byte format )
{
  byte hr;
  byte min;
  byte sec;
  boolean pad0;

  hr   = hour(tm);
  min  = minute(tm);
  sec  = second(tm);
  switch ( format ) {
    case DISPLAY_12HOUR:
      if ( hr > 12 ) {
        hr -= 12;
      }
      digit6 = decode_number( hr  / 10 );
      if( digit6 == Digit_0 ) { digit6 = Digit_NON; }
      digit5 = decode_number( hr  % 10 ) | Digit_DP;
      digit4 = decode_number( min / 10 );
      digit3 = decode_number( min % 10 ) | Digit_DP;
      digit2 = decode_number( sec / 10 );
      digit1 = decode_number( sec % 10 ) | Digit_DP;
      display_refresh();
      break;
    case DISPLAY_24HOUR:
      digit6 = decode_number( hr  / 10 );
      if( digit6 == Digit_0 ) { digit6 = Digit_NON; }
      digit5 = decode_number( hr  % 10 ) | Digit_DP;
      digit4 = decode_number( min / 10 );
      digit3 = decode_number( min % 10 ) | Digit_DP;
      digit2 = decode_number( sec / 10 );
      digit1 = decode_number( sec % 10 ) | Digit_DP;
      display_refresh();
      break;
    case DISPLAY_ALARM:
      digit6 = decode_number( hr  / 10 );
      if( digit6 == Digit_0 ) { digit6 = Digit_NON; }
      digit5 = decode_number( hr  % 10 );
      digit4 = decode_number( min / 10 );
      digit3 = decode_number( min % 10 );
      digit2 = Digit_MINUS;
      digit1 = Digit_MINUS;
      display_refresh();
      break;
    case DISPLAY_DATE:
      digit6 = decode_number( (year(tm) % 100) / 10 );
      digit5 = decode_number( year(tm)  % 10 ) | Digit_DP;
      digit4 = decode_number( month(tm) / 10 );
      digit3 = decode_number( month(tm) % 10 ) | Digit_DP;
      digit2 = decode_number( day(tm)   / 10 );
      digit1 = decode_number( day(tm)   % 10 ) | Digit_DP;
      display_refresh();
      break;
    case DISPLAY_TIMER:
      pad0 = true;
      digit6 = Digit_MINUS;
      digit5 = Digit_MINUS;
      digit4 = decode_number( min / 10 );
      if ( (digit4 == Digit_0) && pad0 ) {
        digit4 = Digit_MINUS;
      } else {
        pad0 = false;
      }
      digit3 = decode_number( min % 10 );
      if ( (digit3 == Digit_0) && pad0 ) {
        digit3 = Digit_MINUS;
      } else {
        pad0 = false;
      }
      digit2 = decode_number( sec / 10 );
      if ( (digit2 == Digit_0) && pad0 ) {
        digit2 = Digit_MINUS;
      } else {
        pad0 = false;
      }
      digit1 = decode_number( sec % 10 );
      display_refresh();
      break;
  }
}

//-------------------------------------------------------------------------------
//  display calibration time
//-------------------------------------------------------------------------------
void display_calibration( int cal_time )
{
  word sign_char;
  byte digit;

  digit6 = Digit_NON;
  digit5 = Digit_NON;
  digit4 = Digit_NON;
  digit3 = Digit_NON;
  digit2 = Digit_NON;
  digit1 = Digit_NON;

  if ( cal_time == 0 ) {
    digit1 = Digit_0;
    display_refresh();
    return;
  }

  if ( cal_time > 0 ) {
    sign_char = Digit_PLUS;
  } else {
    sign_char = Digit_MINUS;
    cal_time = -cal_time;
  }
  digit = 1;
  while ( cal_time > 0 ) {
    switch ( digit ) {
      case 1: digit1 = decode_number( (word)(cal_time % 10) ); break;
      case 2: digit2 = decode_number( (word)(cal_time % 10) ); break;
    }
    cal_time /= 10;
    digit++;
  }
  switch ( digit ) {
    case 2: digit2 = sign_char; break;
    case 3: digit3 = sign_char; break;
  }
  display_refresh();
}

//-------------------------------------------------------------------------------
//  display calibration time
//-------------------------------------------------------------------------------
void display_hour_format( byte format )
{

  digit6 = Digit_NON;
  digit5 = Digit_NON;
  digit4 = Digit_h;
  digit3 = Digit_o;
  digit2 = Digit_u;
  digit1 = Digit_r;

  if( format == DISPLAY_12HOUR ) {
    digit6 = Digit_1;
    digit5 = Digit_2;
  } else {
    digit6 = Digit_2;
    digit5 = Digit_4;
  }
  display_refresh();
}

//-------------------------------------------------------------------------------
//  brink
//-------------------------------------------------------------------------------
void brink( boolean dig6, boolean dig5, boolean dig4, boolean dig3, boolean dig2, boolean dig1 )
{
  word save_dig6;
  word save_dig5;
  word save_dig4;
  word save_dig3;
  word save_dig2;
  word save_dig1;

  save_dig6 = digit6;
  save_dig5 = digit5;
  save_dig4 = digit4;
  save_dig3 = digit3;
  save_dig2 = digit2;
  save_dig1 = digit1;
  for ( byte i = 0 ; i < 3 ; i++ ) {
    if ( dig6 ) { digit6 = 0; }
    if ( dig5 ) { digit5 = 0; }
    if ( dig4 ) { digit4 = 0; }
    if ( dig3 ) { digit3 = 0; }
    if ( dig2 ) { digit2 = 0; }
    if ( dig1 ) { digit1 = 0; }
    display_refresh();
    delay(100);
    digit6 = save_dig6;
    digit5 = save_dig5;
    digit4 = save_dig4;
    digit3 = save_dig3;
    digit2 = save_dig2;
    digit1 = save_dig1;
    display_refresh();
    delay(100);
  }
}

//-------------------------------------------------------------------------------
// input 24hour format time
//-------------------------------------------------------------------------------
time_t input_time( time_t argtm )
{
  tmElements_t tm;
  byte hours;
  byte minutes;

  breakTime( argtm, tm );
  display_time( argtm, DISPLAY_ALARM );
  brink( true, true, false, false, false, false );
  while ( true ) {
    //----- increment hour -------------
    if ( sw_on_test( SW_NO2, SW_DEFAULT_ON_TIME, SW_DEFAULT_OFF_TIME ) ) {
      beep(10);
      if ( tm.Hour == 23 ) {
        tm.Hour = 0;
      } else {
        tm.Hour++;
      }
      display_time( makeTime( tm ), DISPLAY_ALARM );
    }

    //----- decrement hour -------------
    if ( sw_on_test( SW_NO3, SW_DEFAULT_ON_TIME, SW_DEFAULT_OFF_TIME ) ) {
      beep(10);
      if ( tm.Hour == 0 ) {
        tm.Hour = 23;
      } else {
        tm.Hour--;
      }
      display_time( makeTime( tm ), DISPLAY_ALARM );
    }
    if ( sw_on_test( SW_NO1, SW_DEFAULT_ON_TIME, SW_DEFAULT_OFF_TIME) ) {
      break;
    }
  }
  beep(100);

  brink( false, false, true, true, false, false );
  while ( true ) {
    //----- increment minutes -------------
    if ( sw_on_test( SW_NO2, SW_DEFAULT_ON_TIME, SW_DEFAULT_OFF_TIME ) ) {
      beep(10);
      if ( tm.Minute == 59 ) {
        tm.Minute = 0;
      } else {
        tm.Minute++;
      }
      display_time( makeTime( tm ), DISPLAY_ALARM );
    }
    //----- decrement minutes -------------
    if ( sw_on_test( SW_NO3, SW_DEFAULT_ON_TIME, SW_DEFAULT_OFF_TIME ) ) {
      beep(10);
      if ( tm.Minute == 0 ) {
        tm.Minute = 59;
      } else {
        tm.Minute--;
      }
      display_time( makeTime( tm ), DISPLAY_ALARM );
    }
    //------ exit ----------
    if ( sw_on_test( SW_NO1, SW_DEFAULT_ON_TIME, SW_DEFAULT_OFF_TIME) ) {
      break;
    }
  }

  tm.Second = 0;
  beep(100);

  return makeTime( tm );
}

//-------------------------------------------------------------------------------
// input date
//-------------------------------------------------------------------------------
time_t input_date( time_t current_time )
{
  tmElements_t tm;

  display_time( current_time, DISPLAY_DATE );
  breakTime( current_time, tm );

  brink( true, true, false, false, false, false );
  while ( true ) {
    //----- increment year -------------
    if ( sw_on_test( SW_NO2, SW_DEFAULT_ON_TIME, SW_DEFAULT_OFF_TIME ) ) {
      beep(10);
      if ( tm.Year == 2099 ) {
        tm.Year = 0;
      } else {
        tm.Year++;
      }
      display_time( makeTime( tm ), DISPLAY_DATE );
    }
    //----- decrement minutes -------------
    if ( sw_on_test( SW_NO3, SW_DEFAULT_ON_TIME, SW_DEFAULT_OFF_TIME ) ) {
      beep(10);
      if ( tm.Year == 2000 ) {
        tm.Year = 2099;
      } else {
        tm.Year--;
      }
      display_time( makeTime( tm ), DISPLAY_DATE );
    }
    //------ exit ----------
    if ( sw_on_test( SW_NO1, SW_DEFAULT_ON_TIME, SW_DEFAULT_OFF_TIME) ) {
      break;
    }
  }

  brink( false, false, true, true, false, false );
  while ( true ) {
    //----- increment month -------------
    if ( sw_on_test( SW_NO2, SW_DEFAULT_ON_TIME, SW_DEFAULT_OFF_TIME ) ) {
      beep(10);
      if ( tm.Month == 12 ) {
        tm.Month = 1;
      } else {
        tm.Month++;
      }
      display_time( makeTime( tm ), DISPLAY_DATE );
    }
    //----- decrement month -------------
    if ( sw_on_test( SW_NO3, SW_DEFAULT_ON_TIME, SW_DEFAULT_OFF_TIME ) ) {
      beep(10);
      if ( tm.Month == 1 ) {
        tm.Month = 12;
      } else {
        tm.Month--;
      }
      display_time( makeTime( tm ), DISPLAY_DATE );
    }
    //------ break ---------------
    if ( sw_on_test( SW_NO1, SW_DEFAULT_ON_TIME, SW_DEFAULT_OFF_TIME) ) {
      break;
    }
  }
  beep(100);

  brink( false, false, false, false, true, true );
  while ( true ) {
    //----- increment date -------------
    if ( sw_on_test( SW_NO2, SW_DEFAULT_ON_TIME, SW_DEFAULT_OFF_TIME ) ) {
      beep(10);
      if ( tm.Day == 31 ) {
        tm.Day = 1;
      } else {
        tm.Day++;
      }
      display_time( makeTime( tm ), DISPLAY_DATE );
    }
    //----- decrement date -------------
    if ( sw_on_test( SW_NO3, SW_DEFAULT_ON_TIME, SW_DEFAULT_OFF_TIME ) ) {
      beep(10);
      if ( tm.Day == 1 ) {
        tm.Day = 31;
      } else {
        tm.Day--;
      }
      display_time( makeTime( tm ), DISPLAY_DATE );
    }
    //------ exit ----------
    if ( sw_on_test( SW_NO1, SW_DEFAULT_ON_TIME, SW_DEFAULT_OFF_TIME) ) {
      break;
    }
  }
  beep(100);

  return makeTime( tm );
}

//-------------------------------------------------------------------------------
// input calibration time
//-------------------------------------------------------------------------------
void input_calibration()
{
  int calibration_time;

  calibration_time = EEPROM.read( EEPADDR_CALIBRATION_TIME );
  if (  EEPROM.read( EEPADDR_CALIBRATION_SIGN ) == '-' ) {
    calibration_time = -calibration_time;
  }
  display_calibration( calibration_time );
  while ( true ) {
    //----- increment -------------
    if ( sw_on_test( SW_NO2, SW_DEFAULT_ON_TIME, SW_DEFAULT_OFF_TIME ) ) {
      beep(10);
      if ( calibration_time < 99 ) {
        calibration_time++;
      }
      display_calibration( calibration_time );
    }
    //----- decrement -------------
    if ( sw_on_test( SW_NO3, SW_DEFAULT_ON_TIME, SW_DEFAULT_OFF_TIME ) ) {
      beep(10);
      if ( calibration_time > -99 ) {
        calibration_time--;
      }
      display_calibration( calibration_time );
    }
    //------ exit ----------
    if ( sw_on_test( SW_NO1, SW_DEFAULT_ON_TIME, SW_DEFAULT_OFF_TIME) ) {
      if ( calibration_time < 0 ) {
        EEPROM.write( EEPADDR_CALIBRATION_SIGN, '-' );
        EEPROM.write( EEPADDR_CALIBRATION_TIME, (byte)(-calibration_time) );
      } else {
        EEPROM.write( EEPADDR_CALIBRATION_SIGN, '+' );
        EEPROM.write( EEPADDR_CALIBRATION_TIME, (byte)(calibration_time)  );
      }
      crc_eeprom_write();
      break;
    }
  }
}

//-------------------------------------------------------------------------------
// input 12hour / 24hour format
//-------------------------------------------------------------------------------
void input_hour_format()
{
  byte format;
  
  format = EEPROM.read( EEPADDR_DISPLAY_FORMAT );
  display_hour_format( format );

  while ( true ) {
    //----- toggle -------------
    if( sw_on_test( SW_NO2, SW_DEFAULT_ON_TIME, SW_DEFAULT_OFF_TIME ) ||
        sw_on_test( SW_NO3, SW_DEFAULT_ON_TIME, SW_DEFAULT_OFF_TIME ) ) {
      beep(10);
      if( format == DISPLAY_12HOUR ) {
        format = DISPLAY_24HOUR;
      } else {
        format = DISPLAY_12HOUR;
      } 
      display_hour_format( format );
    }
    //------ exit ----------
    if ( sw_on_test( SW_NO1, SW_DEFAULT_ON_TIME, SW_DEFAULT_OFF_TIME) ) {
      EEPROM.write( EEPADDR_DISPLAY_FORMAT, format );
      crc_eeprom_write();
      break;
    }
  }
}  

//-------------------------------------------------------------------------------
// CRC check
//-------------------------------------------------------------------------------
word crc16( word crc, char str[], int len )
{
  const word CRC_CCITT = 0x8408;
  byte i;
  byte j;
  word tmp_word;

  crc = ~crc;
  for ( i = 0 ; i <= (len - 1) ; i++ ) {
    tmp_word = str[i];
    crc = crc ^ tmp_word;
    for ( j = 0 ; j <= 7 ; j++ ) {
      if ( crc & 1 ) {
        crc = (crc >> 1) ^ CRC_CCITT;
      } else {
        crc >>= 1;
      }
    }
  }

  return ~crc;
}

//-------------------------------------------------------------------------------
// HEX to char
//-------------------------------------------------------------------------------
char hex2char( byte hex_value )
{
  if ( hex_value <= 9 ) {
    return hex_value + '0';
  } else {
    return hex_value - 10 + 'A';
  }
}

//-------------------------------------------------------------------------------
// char to HEX
//-------------------------------------------------------------------------------
byte char2hex( char ch )
{
  if ( ch <= '9' ) {
    return ch - '0';
  } else {
    return ch - 'A' + 10;
  }
}

//------------------------------------------------------------------------------------------
// calc eeprom crc
//------------------------------------------------------------------------------------------
word crc_eeprom()
{
  byte eeprom_counter;
  char eep_str[EEPROM_DATA_SIZE];

  for ( eeprom_counter = 0 ; eeprom_counter < EEPROM_DATA_SIZE ; eeprom_counter++ ) {
    eep_str[eeprom_counter] = EEPROM.read( eeprom_counter );
  }

  return crc16( 0xFFFF, eep_str, EEPROM_DATA_SIZE );
}

//------------------------------------------------------------------------------------------
// write eeprom crc
//------------------------------------------------------------------------------------------
void crc_eeprom_write()
{
  word crc_value;

  crc_value = crc_eeprom();
  EEPROM.write( EEPADDR_CRC_LOW,  lowByte(crc_value)  );  // write eeprom crc low byte
  EEPROM.write( EEPADDR_CRC_HIGH, highByte(crc_value) );  // write eeprom crc high byte
}

//------------------------------------------------------------------------------------------
// initialize eeprom
//------------------------------------------------------------------------------------------
void initialize_eeprom()
{
  word crc;

  EEPROM.write( EEPADDR_ALARM_HOURS,      ALARM_DEFAULT_HOUR   );
  EEPROM.write( EEPADDR_ALARM_MINUTES,    ALARM_DEFAULT_MINUTE );
  EEPROM.write( EEPADDR_ALARM_ENABLE,     ALARM_ENABLE );
  EEPROM.write( EEPADDR_CALIBRATION_SIGN, '+' );
  EEPROM.write( EEPADDR_CALIBRATION_TIME, DEFAULT_CALIBRATION_TIME );
  EEPROM.write( EEPADDR_DISPLAY_FORMAT, DISPLAY_24HOUR );
  EEPROM.write( EEPADDR_TIMER_SEC, 30 );
  EEPROM.write( EEPADDR_TIMER_MIN, 0  );

  crc_eeprom_write();
}

//------------------------------------------------------------------------------------------
// kitchen timer
//------------------------------------------------------------------------------------------
void kitchen_timer()
{
  word    timer_seconds;
  word    timer_minutes;
  boolean start_timer;
  time_t  timer_initial_value;
  time_t  timer_count;

  timer_seconds = EEPROM.read( EEPADDR_TIMER_SEC );
  timer_minutes = EEPROM.read( EEPADDR_TIMER_MIN );
  timer_initial_value = timer_minutes * 60 + timer_seconds;
  display_time( timer_initial_value, DISPLAY_TIMER );
  start_timer = false;
  while ( true ) {
    //----- decrement timer -------------
    if ( sw_on_test( SW_NO3, SW_DEFAULT_ON_TIME, SW_DEFAULT_OFF_TIME ) ) {
      if ( !start_timer ) {
        if ( timer_initial_value > 10 ) {
          timer_initial_value -= 10;
          display_time( timer_initial_value, DISPLAY_TIMER );
          beep(10);
        }
      }
    }
    //----- increment timer -------------
    if ( sw_on_test( SW_NO2, SW_DEFAULT_ON_TIME, SW_DEFAULT_OFF_TIME ) ) {
      if ( !start_timer ) {
        if ( timer_initial_value < 3600 ) {
          timer_initial_value += 10;
          display_time( timer_initial_value, DISPLAY_TIMER );
          beep(10);
        }
      }
    }
    //----- start timer -------------
    if ( sw_on_test( SW_NO1, SW_DEFAULT_ON_TIME, SW_DEFAULT_OFF_TIME ) ) {
      if ( start_timer ) {
        start_timer = false;
        display_time( timer_initial_value, DISPLAY_TIMER );
        beep2(2);
      } else {
        timer_count = timer_initial_value;
        start_timer = true;
        beep(100);
        kitchen_counter_tick = false;
        kitchen_counter_init = true;
      }
    }
    //----- count doun -------------
    if ( kitchen_counter_tick == true ) {
      if ( start_timer ) {
        if ( timer_count > 0 ) {
          timer_count--;
        }
        if ( timer_count == 0 ) {
          display_time( timer_count, DISPLAY_TIMER );
          beep2(10);
          start_timer = false;
          delay(1000);
          timer_count = timer_initial_value;
        }
        display_time( timer_count, DISPLAY_TIMER );
      }
      kitchen_counter_tick = false;
    }
    //------ exit ----------
    if ( sw_on_test( SW_NO4, SW_DEFAULT_ON_TIME, SW_DEFAULT_OFF_TIME) ) {
      beep(100);
      break;
    }
  }
}

//------------------------------------------------------------------------------------------
// arduino setup
//------------------------------------------------------------------------------------------
void setup()
{
  time_t tm;
  time_t power_off_time;
  unsigned long diff_sec;
  tmElements_t alarm_time_struct;
  word eeprom_crc_value;

  pinMode( BUZZER, OUTPUT );
  pinMode( LED,    OUTPUT );
  pinMode( RTC_OUT, INPUT );
  pinMode( SW1, INPUT );
  pinMode( SW2, INPUT );
  pinMode( SW3, INPUT );
  pinMode( SW4, INPUT );
  pinMode( VOLTAGE_DETECTOR, INPUT );
  pinMode( DC_DC_DISABLE, OUTPUT );
  pinMode( A6812_CLOCK,  OUTPUT );    // VFD 20-bit serial-input latched source driver
  pinMode( A6812_DATA ,  OUTPUT );
  pinMode( A6812_BLANK,  OUTPUT );
  pinMode( A6812_STROBE, OUTPUT );

  digitalWrite( A6812_CLOCK,  LOW  );
  digitalWrite( A6812_BLANK,  HIGH );
  digitalWrite( A6812_STROBE, LOW  );
  digitalWrite( DC_DC_DISABLE, LOW  );

  Wire.begin();
  display_clear();

  sw1_enable = true;
  sw2_enable = true;
  sw3_enable = true;
  sw4_enable = true;

  alarm_beep_enable      = false;
  alarm_beep_counter     = 0;
  alarm_beep_table_index = 0;

  eeprom_crc_value = 0;
  eeprom_crc_value = EEPROM.read( EEPADDR_CRC_HIGH );
  eeprom_crc_value <<= 8;
  eeprom_crc_value += EEPROM.read( EEPADDR_CRC_LOW );
  if ( eeprom_crc_value != crc_eeprom() ) {
    initialize_eeprom();
  }
  // initialize_eeprom();

  Serial.begin(9600);

  write_ds1307( ADDRESS_CONTROL, OUT_1Hz);
  if ( ds1307_get_halt() ) {
    ds1307_settime( 17, 1, 1, 0, 0, 0 );
    beep_error();
  }

  //--- enable interrupt ------------------------
  attachInterrupt(0, rtcEvent, CHANGE );  // Interrupt on falling edge
  MsTimer2::set(1, timerEvent );  // 1[msec] period
  MsTimer2::start();

  //--------- time set ----------------------------------------
  if ( sw_on_test(SW_NO1, SW_DEFAULT_ON_TIME, SW_DEFAULT_OFF_TIME) ) {
    beep2(2);
    tm = input_date( ds1307_gettime() );
    tm = input_time( tm );
    input_hour_format();
    settime( tm );  // set DS1307
    input_calibration();
  }

  alarm_time_struct.Hour   = EEPROM.read( EEPADDR_ALARM_HOURS );
  alarm_time_struct.Minute = EEPROM.read( EEPADDR_ALARM_MINUTES );
  alarm_time = makeTime( alarm_time_struct );
  alarm_enable = get_alarm_enable();

  cal_sign = EEPROM.read( EEPADDR_CALIBRATION_SIGN );
  cal_time = EEPROM.read( EEPADDR_CALIBRATION_TIME );
  cal_set = 86400 / cal_time;
  cal_counter = 0;
  display_format = EEPROM.read( EEPADDR_DISPLAY_FORMAT );

  gps_buffer_index = 0;
  gps_buffer_full  = false;

  tick_flag = false;
  beep(100);

  current_time = ds1307_gettime();
  power_off_time = ds1307_read_ram_long( OFFSET_ADDRESS_POWER_OFF_TIME );
  if( power_off_time != 0 ) {
    diff_sec = current_time - power_off_time;
    if( diff_sec <= ( 7 * 86400 ) ) {
      if( cal_sign == '+' ) {
        current_time -= (time_t)(diff_sec / cal_set); 
      } else {
        current_time += (time_t)(diff_sec / cal_set); 
      }
      settime( current_time );      
    }
    cal_counter = ds1307_read_ram_long( OFFSET_ADDRESS_CALIBRATION_COUNTER );
  } else {
    beep_error();
  }  
  
  ds1307_write_ram_long( OFFSET_ADDRESS_POWER_OFF_TIME,      0 );
  ds1307_write_ram_long( OFFSET_ADDRESS_CALIBRATION_COUNTER, 0 );
}

//------------------------------------------------------------------------------------------
// arduino loop
//------------------------------------------------------------------------------------------
void loop()
{
  tmElements_t alarm_time_struct;  
  tmElements_t gps_time_struct;  

  time_t        gps_time;
  byte tmp_byte;
  word crc_HW;
  word crc_PC;
  byte token_index;
  byte buffer_index;

  //--------- time update ---------------------------------------
  if ( tick_flag == true ) {
    //------ display update -------------------------------
    display_time( current_time, display_format );
    //------ alarm ----------------------------------------
    if ( alarm_enable ) {
      if ( alarm_beep_enable == false ) {
        if ( ( hour(alarm_time) == hour(current_time) ) && ( minute(alarm_time) == minute(current_time) ) && ( second(current_time) == 0 ) ) {
          alarm_beep_on();
        }
      } else {
        if ( ( hour(alarm_time) != hour(current_time) ) || ( minute(alarm_time) != minute(current_time) ) ) {
          alarm_beep_off();
        }
      }
    }
    //------ calibration -----------------------------------
    if ( cal_time != 0 ) {
      cal_counter++;
      if ( cal_counter >= cal_set ) {
        if ( cal_sign == '+' ) {
          current_time--;
        } else {
          current_time++;
        }
        settime( current_time );
        cal_counter = 0;
      }
    }
    tick_flag = false;
  }

  //--------- power off -----------------------------------
  if ( digitalRead( VOLTAGE_DETECTOR ) == 0 ) {
    digitalWrite( DC_DC_DISABLE, HIGH  );
    led_on();
    ds1307_write_ram_long( OFFSET_ADDRESS_POWER_OFF_TIME,      current_time );
    ds1307_write_ram_long( OFFSET_ADDRESS_CALIBRATION_COUNTER, cal_counter  );
    write_ds1307( ADDRESS_CONTROL, OUT_OFF );
    led_off();
    while( true ){;}
  }

  //--------- alarm beep off -----------------------------------
  if ( sw_on_test(SW_NO3, SW_DEFAULT_ON_TIME, 500) ) {
    if ( alarm_beep_enable ) {
      alarm_beep_off();
    } else {
      beep(100);
      display_time( current_time, DISPLAY_DATE );
      delay(3000);
      beep(20);
    }
  }

  //--------- alarm enable / disable -----------------------------------
  if ( sw_on_test(SW_NO2, SW_DEFAULT_ON_TIME, 500) ) {
    alarm_beep_off();
    if ( alarm_enable ) {
      beep(100);
      alarm_enable = false;
    } else {
      beep2(4);
      alarm_enable = true;
    }
    put_alarm_enable( alarm_enable );
  }

  //--------- alarm display & alarm time set -----------------------------------
  if ( sw_on_test(SW_NO1, SW_DEFAULT_ON_TIME, 500) ) {
    alarm_beep_off();
    beep(100);
    display_time( alarm_time, DISPLAY_ALARM );
    word time_remain = 3000;
    while ( time_remain > 0 ) {
      if ( sw_on_test(SW_NO1, SW_DEFAULT_ON_TIME, SW_DEFAULT_OFF_TIME) ) {
        beep2(4);
        alarm_time = input_time( alarm_time );
        EEPROM.write( EEPADDR_ALARM_HOURS,   hour( alarm_time )   );
        EEPROM.write( EEPADDR_ALARM_MINUTES, minute( alarm_time ) );
        crc_eeprom_write();
        break;
      }
      delay(1);
      time_remain--;
    }
    beep(20);
    display_time( ds1307_gettime(), display_format );
  }

  //--------- kitchen timer -----------------------------------
  if ( sw_on_test(SW_NO4, SW_DEFAULT_ON_TIME, 500) ) {
    beep(100);
    kitchen_timer();
  }

  //--------- Serial Port -----------------------------------
  // example
  // GPS:  $GPRMC,101105.00,A,3552.52403,N,13917.10887,E,0.057,,170915,,,D*72
  // USB Serial:
  //    ? <- Query
  //    !13,22,07:00,E,+10 <- brightness & alarm & calibration setup
  if( gps_buffer_full ) {
    if( gps_buffer[0] == '$' ) {
      switch( gps_buffer[1] ) {
        case '#':
          switch( gps_buffer[2] ) {
            case '?':   // recieve Query command
              beep(10);
              serial_write_buffer[0] = 'R';
              for( byte index = 1 ; index <= 6 ; index++ ) {
                serial_write_buffer[index] = ' ';
              }
              tmp_byte = EEPROM.read( EEPADDR_ALARM_HOURS  );
              serial_write_buffer[7] = (tmp_byte /   10) + '0';
              serial_write_buffer[8] = (tmp_byte % 10) + '0';
              serial_write_buffer[9] = ':';
              tmp_byte = EEPROM.read( EEPADDR_ALARM_MINUTES  );
              serial_write_buffer[10] = (tmp_byte /   10) + '0';
              serial_write_buffer[11] = (tmp_byte % 10) + '0';
              serial_write_buffer[12] = ',';
              if( get_alarm_enable == ALARM_ENABLE ) {
                serial_write_buffer[13] = 'E';
              } else {
                serial_write_buffer[13] = 'D';
              }
              serial_write_buffer[14] = ',';
              serial_write_buffer[15] = EEPROM.read( EEPADDR_CALIBRATION_SIGN );
              tmp_byte = EEPROM.read( EEPADDR_CALIBRATION_TIME  );
              serial_write_buffer[16] = (tmp_byte /   10) + '0';
              serial_write_buffer[17] = (tmp_byte % 10) + '0';
              serial_write_buffer[18] = ' ';
              crc_HW = crc16(0xFFFF, serial_write_buffer, 18);
              serial_write_buffer[22] = hex2char(crc_HW % 0x10);
              crc_HW = crc_HW >> 4;
              serial_write_buffer[21] = hex2char(crc_HW % 0x10);
              crc_HW = crc_HW >> 4;
              serial_write_buffer[20] = hex2char(crc_HW % 0x10);
              crc_HW = crc_HW >> 4;
              serial_write_buffer[19] = hex2char(crc_HW % 0x10);
              serial_write_buffer[23] = 0;
              Serial.println( serial_write_buffer );
              break;
            case 'C':    // recieve Change command
              crc_HW = crc16(0xFFFF, gps_buffer, 20);
              crc_PC = 0;
              crc_PC = crc_PC + char2hex( gps_buffer[21] );  crc_PC = crc_PC << 4;
              crc_PC = crc_PC + char2hex( gps_buffer[22] );  crc_PC = crc_PC << 4;
              crc_PC = crc_PC + char2hex( gps_buffer[23] );  crc_PC = crc_PC << 4;
              crc_PC = crc_PC + char2hex( gps_buffer[24] );

              if( crc_HW == crc_PC ) {
                beep(10);
                alarm_time_struct.Hour   = (gps_buffer[9]  - '0') * 10 + (gps_buffer[10] - '0');
                alarm_time_struct.Minute = (gps_buffer[12] - '0') * 10 + (gps_buffer[13] - '0');
                alarm_time_struct.Second = 0;
                alarm_time = makeTime( alarm_time_struct );
                EEPROM.write( EEPADDR_ALARM_HOURS,   alarm_time_struct.Hour   );
                EEPROM.write( EEPADDR_ALARM_MINUTES, alarm_time_struct.Minute );
                if( gps_buffer[15] == 'D' ) {
                  alarm_enable = false;
                } else {
                  alarm_enable = true;
                }
                put_alarm_enable( alarm_enable );
                cal_sign = gps_buffer[17];
                cal_time = (gps_buffer[18] - '0') * 10 + (gps_buffer[19] - '0');
                cal_set = 86400 / cal_time;
                cal_counter = 0;
                EEPROM.write( EEPADDR_CALIBRATION_SIGN, cal_sign );
                EEPROM.write( EEPADDR_CALIBRATION_TIME, cal_time );
                crc_eeprom_write();
              }
              break;
          }
          break;
        case 'G':    // recieve GPS NMEA format
          if ( (gps_buffer[2] == 'P') && (gps_buffer[3] == 'R') && (gps_buffer[4] == 'M') && (gps_buffer[5] == 'C') && (gps_buffer[7] != ',') ) {
            gps_time_struct.Hour   = (gps_buffer[7]  - 0x30) * 10 + (gps_buffer[8]  - 0x30);
            gps_time_struct.Minute = (gps_buffer[9]  - 0x30) * 10 + (gps_buffer[10] - 0x30);
            gps_time_struct.Second = (gps_buffer[11] - 0x30) * 10 + (gps_buffer[12] - 0x30);
            token_index = 0;
            for ( buffer_index = 0 ; buffer_index <= 64 ; buffer_index++ ) {
              if ( gps_buffer[buffer_index] == ',' ) {
                token_index++;
                if ( token_index == 9 ) {
                  break;
                }
              }
            }
            if( gps_buffer[buffer_index + 1] != ',' ) {
              gps_time_struct.Day   = (gps_buffer[buffer_index + 1] - 0x30) * 10 + (gps_buffer[buffer_index + 2] - 0x30);
              gps_time_struct.Month = (gps_buffer[buffer_index + 3] - 0x30) * 10 + (gps_buffer[buffer_index + 4] - 0x30);
              gps_time_struct.Year  = (gps_buffer[buffer_index + 5] - 0x30) * 10 + (gps_buffer[buffer_index + 6] - 0x30) + 30;
              gps_time = makeTime( gps_time_struct );
              gps_time += 32400;   // 32400[sec] = 9[hours]   GMT+9
              if( (current_time < (gps_time - 1)) || ((gps_time + 1) < current_time) ) {
                current_time = gps_time;
                settime( current_time );
                beep(100);
              }
            }
          }
          break;
      }
    }
    gps_buffer[0] = ' ';
    gps_buffer_full = false;
  }
}

//---------------------------------------------------- E N D ------------------------------------------------
